import 'dart:async';
import 'dart:convert';

import 'package:http/http.dart' as http;
import 'package:http/http.dart';
import 'package:meta/meta.dart';
import 'package:osam_http_client/src/exceptions.dart';

String bearer(String token) => 'bearer $token';

abstract class HttpClient {
  factory HttpClient(
          {Duration timeout = const Duration(seconds: 10), bool log = false}) =>
      _SimpleHttpClient(timeout: timeout, log: log);

  Future<Response> get({
    @required String host,
    @required String path,
    Map<String, String> query,
    Map<String, String> headers,
  });

  Future<Response> post({
    @required String host,
    @required String path,
    Map<String, String> query,
    Map<String, String> headers,
    Object body,
  });
}

class _SimpleHttpClient implements HttpClient {
  final Duration timeout;
  final bool log;

  _SimpleHttpClient({this.timeout, this.log});

  Map<String, dynamic> _checkResponseBody(Response response) {
    final body = response.body;
    if (body.isNotEmpty && body != null) {
      return jsonDecode(body);
    }
    return <String, dynamic>{};
  }

  Response _checkedResponse(Response response) {
    final statusCode = response.statusCode;
    if (statusCode >= 200 && statusCode < 300) return response;
    if (statusCode == 401)
      throw HttpUnauthorizedException(_checkResponseBody(response));
    if (statusCode >= 400 && statusCode < 500)
      throw HttpClientException(_checkResponseBody(response));
    if (statusCode >= 500 && statusCode < 600)
      throw HttpServerException(_checkResponseBody(response));
    throw HttpUnknownException(_checkResponseBody(response));
  }

  @override
  Future<Response> get({
    @required String host,
    @required String path,
    Map<String, String> query,
    Map<String, String> headers,
  }) {
    final queryString = _makeQuery(query);
    final fullPath = '$host$path$queryString';
    if (log) print('path: $fullPath,\nheaders: $headers');
    return http
        .get(Uri.encodeFull(fullPath), headers: headers)
        .timeout(timeout)
        .catchError((Object error) {
      if (log) print(error);
      throw NoConnectionException();
    }).then((response) {
      if (log) print(response.log());
      return _checkedResponse(response);
    });
  }

  @override
  Future<Response> post({
    @required String host,
    @required String path,
    Map<String, String> query,
    Map<String, String> headers,
    Object body,
  }) {
    final queryString = _makeQuery(query);
    final fullPath = '$host$path$queryString';
    if (log) print('path: $fullPath,\nheaders: $headers, \nbody');
    return http
        .post(Uri.encodeFull(fullPath), headers: headers, body: body)
        .timeout(timeout)
        .catchError((Object error) {
      if (log) print(error);
      throw NoConnectionException();
    }).then((response) {
      if (log) print(response.log());
      return _checkedResponse(response);
    });
  }
}

extension Log on Response {
  String log() {
    return 'headers: $headers\nstatusCode: $statusCode\nbody: $body';
  }
}

String _makeQuery(Map<String, String> queryParameters) {
  if (queryParameters == null || queryParameters.isEmpty) {
    return '';
  }
  final result = StringBuffer('?');
  var separator = '';
  void writeParameter(String key, String value) {
    result.write(separator);
    separator = '&';
    result.write(key);
    if (value != null && value.isNotEmpty) {
      result.write('=');
      result.write(value);
    }
  }

  queryParameters.forEach((key, value) {
    if (value != null) {
      writeParameter(key, value);
    }
  });
  return result.toString();
}
